/* Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <pnfft.h>
#include <ipnfft.h>

static void fft_output_size(
    const INT *n, const R *x_max, int m,
    INT *no);


void PNX(local_size_guru)(
    int d, const INT *N, const INT *n, const R *x_max, int m,
    MPI_Comm comm_cart_2d,
    INT *local_N, INT *local_N_start,
    R *lower_border, R *upper_border
    )
{
  INT no[3], local_no[3], local_no_start[3];

  if(d != 3){
    PNX(message)("!!! Error in PNFFT: d != 3 not yet implemented !!!\n", comm_cart_2d);
    return;
  }

  fft_output_size(n, x_max, m,
      no);

  PNX(local_size_internal)(N, n, no, comm_cart_2d,
      local_N, local_N_start, local_no, local_no_start);

  PNX(node_borders)(n, no, local_no, local_no_start,
      lower_border, upper_border);
}


void PNX(init_guru)(
    PNX(plan) *ths, int d, const INT *N, const INT *n, const R *x_max,
    INT local_M, int m,
    unsigned nfft_flags, unsigned fftw_flags,
    MPI_Comm comm_cart_2d
    )
{
  INT no[3];

  ths->d = d;
  ths->m= m;
  ths->local_M = local_M;

  fft_output_size(n, x_max, m,
      no);

  ths->N = (INT*) PNX(malloc)(d * sizeof(INT), comm_cart_2d);
  ths->n = (INT*) PNX(malloc)(d * sizeof(INT), comm_cart_2d);
  ths->no= (INT*) PNX(malloc)(d * sizeof(INT), comm_cart_2d);
  for(int t=0; t<d; t++){
    ths->N[t]= N[t];
    ths->n[t]= n[t];
    ths->no[t]= no[t];
  }


  ths->nfft_flags = nfft_flags;
  ths->fftw_flags = fftw_flags;

  MPI_Comm_dup(comm_cart_2d, &(ths->comm_cart_2d));

  PNX(init_help)(ths);
}


static void fft_output_size(
    const INT *n, const R *x_max, int m,
    INT *no
    )
{
  INT c;

  for(int t=0; t<3; t++){
    c = (INT) floor(n[t]*x_max[t]);
    no[t] = MIN(n[t], 2*(c+m+2));
  }
}


