/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* This file is based on ifftw3.h and ifftw3-mpi.h */


/* PFFT internal header file */
#ifndef __IPNFFT_H__
#define __IPNFFT_H__


/* Begin: This part is based on ifftw3.h */
#include <stdlib.h>             /* size_t */
#include <stdarg.h>             /* va_list */
#include <stddef.h>             /* ptrdiff_t */
#include <stdio.h>              /* fprintf */
#include <memory.h>             /* memset */

#include <mpi.h>
#include <fftw3-mpi.h>
#include <pfft.h>

#define IPNFFT_EXTERN extern

typedef ptrdiff_t INT;
/*
  integral type large enough to contain a stride (what ``int'' should
  have been in the first place.
*/

#define IF(x,a,b) ((x)?(a):(b))

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))
#define ABS(x) (((x)>0)?(x):(-(x)))
#define SIGN(a) (((a)>=0)?1:-1)
#define SQR(a) ((a)*(a))

#define MACRO_PLAIN_INDEX_3D(k, n)      ( k[2] + n[2]*(k[1] + n[1]*k[0]) )
#define MACRO_FFTSHIFT(k, N)            ( ((N)/2-1 < (k)) ? (k)-(N) : (k) )
#define MACRO_INV_PHI_HAT(k, n, b)         ( exp( M_PI*M_PI*(k)*(k)*1.0/((n)*(n))*(b)) )
//#define PLAIN_INDEX(k0,k1,k2,n)           ((k2)+n[2]*((k1)+n[1]*(k0)))

/* GAUSSIAN window function */
#define PHI_HUT(k,d) ((double)exp(-(pow(PI*(k)/ths->n[d],2.0)*ths->b[d])))
#define PHI(x,d) ((double)exp(-pow((x)*ths->n[d],2.0)/                         \
                   ths->b[d])/sqrt(PI*ths->b[d]))


/* all pnfft identifiers start with pnfft (or pnfftf etc.) */
#define CONCAT(prefix, name) prefix ## name
#if defined(PNFFT_SINGLE)
  typedef float R;
  typedef fftwf_complex C;
#  define PNFFT_MPI_TYPE MPI_FLOAT
#  define PNX(name) CONCAT(pnfftf_, name)
#  define PX(name) CONCAT(pfftf_, name)
#  define X(name)  CONCAT(fftwf_, name)
#elif defined(PNFFT_LDOUBLE)
  typedef long double R;
  typedef fftwl_complex C;
#  define PNFFT_MPI_TYPE MPI_LONG_DOUBLE
#  define PNX(name) CONCAT(pnfftl_, name)
#  define PX(name) CONCAT(pfftl_, name)
#  define X(name)  CONCAT(fftwl_, name)
#else
  typedef double R;
  typedef fftw_complex C;
#  define PNFFT_MPI_TYPE MPI_DOUBLE
#  define PNX(name) CONCAT(pnfft_, name)
#  define PX(name) CONCAT(pfft_, name)
#  define X(name)  CONCAT(fftw_, name)
#endif

#define XM(name)  X(CONCAT(mpi_, name))

/* utils.c */
void PNX(message)(
    const char *string, MPI_Comm comm);
INT PNX(prod_ptrdiff_t)(
    int d, const INT *vec);
int PNX(equal_ptrdiff_t)(
    int d, const INT *vec1, const INT *vec2);
void PNX(vcopy_ptrdiff_t)(
    int d, const INT *vec1,
    INT *vec2);
void PNX(vadd_ptrdiff_t)(
    int d, const INT *vec1, const INT *vec2,
    INT *sum);
void PNX(vsub_ptrdiff_t)(
    int d, const INT *vec1, const INT *vec2,
    INT *sum);

/* malloc.c */
void PNX(die)(
    const char *s, MPI_Comm comm);


/* ndft-parallel.c */
void PNX(local_size_D)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart_2d,
    INT *local_N, INT *local_N_start);
void PNX(local_size_B)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart_2d,
    INT *local_no, INT *local_no_start);
void PNX(local_size_internal)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart_2d,
    INT *local_N, INT *local_N_start,
    INT *local_no, INT *local_no_start);
void PNX(init_help)(
    PNX(plan) *ths);
void PNX(trafo_D)(
    PNX(plan) *ths);
void PNX(adjoint_D)(
    PNX(plan) *ths);
void PNX(trafo_F)(
    PNX(plan) *ths);
void PNX(adjoint_F)(
    PNX(plan) *ths);
void PNX(trafo_B)(
    PNX(plan) *ths);
void PNX(adjoint_B)(
    PNX(plan) *ths);
void PNX(init_x_and_f)(
    PNX(plan) *ths,
    unsigned nfft_flags);
void PNX(node_borders)(
    const INT *n, const INT *no,
    const INT* local_no, const INT *local_no_start,
    R *lo, R *up);


#endif /* __IPNFFT_H__ */
